/*	$Id$	*/

/*
 * Copyright (c) 2012, Internet Initiative Japan, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <inttypes.h>
#include <string.h>
#include <sys/ioctl.h>

#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/rand.h>

#include <libarms_log.h>
#include <libarms_resource.h>
#include <transaction/transaction.h>
#include <libarms/ssl.h>

static X509 *ssl_mycert;
static EVP_PKEY *ssl_mykey;

/*0:LS 1:RS*/
static X509 *ssl_cacert;

static void free_certificate(void);
static int verify_ls_cn(X509_STORE_CTX *, void *);
static int arms_ssl_dhparam(SSL_CTX *);
#ifdef USE_ECDH
static int arms_ssl_ecdh_curve(SSL_CTX *, const char *);
#endif

void
arms_ssl_init(void)
{
	SSL_load_error_strings();
	SSL_library_init();
}

int
arms_ssl_register_cacert(const char *cacert)
{
	BIO *mem;
	X509 *tmp_cacert;

	ERR_clear_error();

	mem = BIO_new_mem_buf((void *)cacert, -1);
	tmp_cacert = PEM_read_bio_X509(mem, NULL, NULL, NULL);
	BIO_free(mem);
	if (tmp_cacert == NULL) {
		return -1;
	}
	if (ssl_cacert != NULL) {
		X509_free(ssl_cacert);
	}
	ssl_cacert = tmp_cacert;
	return 0;
}

/*
 * originally from http/http.c::http_register_cert()
 */
int
arms_ssl_register_cert(const char *mycert, const char *mykey)
{
	BIO *mem;
	EVP_PKEY *tmp_mykey;
	X509 *tmp_mycert;
	int error = 0;

	tmp_mycert = NULL;
	tmp_mykey = NULL;

	ERR_clear_error();

	if (mycert != NULL && strlen(mycert) > 0) {
		mem = BIO_new_mem_buf((void *)mycert, -1);
		tmp_mycert = PEM_read_bio_X509(mem, NULL, NULL, NULL);
		BIO_free(mem);

		if (tmp_mycert == NULL) {
			error = 1;
			goto err;
		}
	}

	if (mykey != NULL && strlen(mykey) > 0) {
		mem = BIO_new_mem_buf((void *)mykey, -1);
		tmp_mykey = PEM_read_bio_PrivateKey(mem, NULL, NULL, NULL);
		BIO_free(mem);

		if (tmp_mykey == NULL) {
			error = 2;
			goto err;
		}
	}


	free_certificate();
	ssl_mycert = tmp_mycert;
	ssl_mykey = tmp_mykey;

	return 0;

err:
	if (tmp_mycert != NULL)
		X509_free(tmp_mycert);
	if (tmp_mykey != NULL)
		EVP_PKEY_free(tmp_mykey);
	libarms_log(ARMS_LOG_ECERTIFICATE,
		    "Registering certification got error.");
	return error;
}

X509 *
arms_ssl_mycert(void)
{
	return ssl_mycert;
}
EVP_PKEY *
arms_ssl_mykey(void)
{
	return ssl_mykey;
}

X509 *
arms_ssl_cacert(void)
{
	return ssl_cacert;
}

static void
free_certificate(void)
{
	if (ssl_mycert != NULL) {
		X509_free(ssl_mycert);
		ssl_mycert = NULL;
	}
	if (ssl_mykey != NULL) {
		EVP_PKEY_free(ssl_mykey);
		ssl_mykey = NULL;
	}
}

int
arms_ssl_servercert_verify_cb(int ok, X509_STORE_CTX *ctx)
{
	char cn[256];
	const char *errmsg = NULL;

	X509_NAME_oneline(
		X509_get_subject_name(
			X509_STORE_CTX_get_current_cert(ctx)), cn, sizeof(cn));

	X509_NAME_oneline(
		X509_get_issuer_name(
			X509_STORE_CTX_get_current_cert(ctx)), cn, sizeof(cn));

	if (!ok) {
		switch (ctx->error) {
		case X509_V_ERR_CERT_NOT_YET_VALID:
		case X509_V_ERR_CERT_HAS_EXPIRED:
			/* XXX: ignore Validity Not Before/Not After field */
			ok = 1;
			ctx->error = X509_V_OK;
			break;

		case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
			errmsg =
				"unable to get local issuer certificate"
				"(certificate chain may be too long)";
			break;

		default:
			errmsg = X509_verify_cert_error_string(ctx->error);
			break;
		}

	}
	if (ok) {
		/* call libarms verify function */
		SSL *ssl;
		transaction *tr;

		ssl = X509_STORE_CTX_get_ex_data(
			ctx,
			SSL_get_ex_data_X509_STORE_CTX_idx());

		/* tunnel ssl is no transaction required. */
		tr = SSL_get_ex_data(ssl, 0);
		if (tr != NULL && TR_TYPE(tr->state) == TR_LSPULL)
			ok = (verify_ls_cn(ctx, tr) == X509_V_OK);
	}
	if (!ok) {
		libarms_log(ARMS_LOG_ESSL,
			    "verification failure of server certificate");
		libarms_log(ARMS_LOG_ESSL,
			    "reason: %s", errmsg);
	}
	return ok;
}

static int
verify_ls_cn(X509_STORE_CTX *x509ctx, void *u)
{
	char subject[256];

	X509_NAME_oneline(
		X509_get_subject_name(
			X509_STORE_CTX_get_current_cert(x509ctx)),
		subject, sizeof(subject));

	if (strstr(subject, "CN=ARMS Root CA") != NULL)
		return X509_V_OK;

	if (strstr(subject, "CN=Location Server ") != NULL)
		return X509_V_OK;

	return -1;
}

void
arms_ssl_register_randomness(const void *seed, unsigned int len)
{
	unsigned long zero = 0;
	int n;
	const int limit = 10000;

	if (seed != NULL && len != 0)
		RAND_seed(seed, len);
	for (n = 0; RAND_status() != 1 && n < limit; n++)
		RAND_seed(&zero, sizeof(zero));
	/* if (n == limit) give up */
}

static void
arms_log_ssl_error(void)
{
	char errbuf[128];
	int code, line, flags;
	const char *file, *data;

	while ((code = ERR_get_error_line_data(&file, &line, &data, &flags)) != 0) {
		ERR_error_string_n(code, errbuf, sizeof(errbuf) - 8);
		libarms_log(ARMS_LOG_ESSL, errbuf);
		if (data != NULL && (flags & ERR_TXT_STRING)) {
			libarms_log(ARMS_LOG_ESSL, data);
		}
	}
}

static int
arms_ssl_dhparam(SSL_CTX *ctx)
{
	DH *dh;

	/*
	 * -----BEGIN DH PARAMETERS-----
	 *  MIIBCAKCAQEA3kz48kDaPqejY70f9tO2Ncjmk9f/+kC3rZ/xM+PBdKMwK3YceaBx
	 *  cLmEJPygST4Q/je+QRng0+BzO79wODaZox+okcXYKu01MhGaa/X34FYX+vhD7N8/
	 *  XxVB9FXoHn8zn0q6pc2lJRP5T44R68pz6WL0wXsQvlrsKi/azufoyqvCbWmSjNWN
	 *  m4VEB9uYGVvb0UsSPZpWYSRKKXLmHDlm/kFiUW1RrZufrTB0dgtPprtN5tUrEisE
	 *  2c4EAKCuPzPcRulAvISaMMr6YaV8Ismna1Lk1YoWYljlY62ce5nxUbZ/HTPDq97k
	 *  si7lULr4A7Kx6vSqj03n5Zh4RWM2/vhq8wIBAg==
	 *  -----END DH PARAMETERS-----
	 */

	static unsigned char dh2048_p[] = {
	    0xde, 0x4c, 0xf8, 0xf2, 0x40, 0xda, 0x3e, 0xa7, 0xa3, 0x63, 0xbd,
	    0x1f, 0xf6, 0xd3, 0xb6, 0x35, 0xc8, 0xe6, 0x93, 0xd7, 0xff, 0xfa,
	    0x40, 0xb7, 0xad, 0x9f, 0xf1, 0x33, 0xe3, 0xc1, 0x74, 0xa3, 0x30,
	    0x2b, 0x76, 0x1c, 0x79, 0xa0, 0x71, 0x70, 0xb9, 0x84, 0x24, 0xfc,
	    0xa0, 0x49, 0x3e, 0x10, 0xfe, 0x37, 0xbe, 0x41, 0x19, 0xe0, 0xd3,
	    0xe0, 0x73, 0x3b, 0xbf, 0x70, 0x38, 0x36, 0x99, 0xa3, 0x1f, 0xa8,
	    0x91, 0xc5, 0xd8, 0x2a, 0xed, 0x35, 0x32, 0x11, 0x9a, 0x6b, 0xf5,
	    0xf7, 0xe0, 0x56, 0x17, 0xfa, 0xf8, 0x43, 0xec, 0xdf, 0x3f, 0x5f,
	    0x15, 0x41, 0xf4, 0x55, 0xe8, 0x1e, 0x7f, 0x33, 0x9f, 0x4a, 0xba,
	    0xa5, 0xcd, 0xa5, 0x25, 0x13, 0xf9, 0x4f, 0x8e, 0x11, 0xeb, 0xca,
	    0x73, 0xe9, 0x62, 0xf4, 0xc1, 0x7b, 0x10, 0xbe, 0x5a, 0xec, 0x2a,
	    0x2f, 0xda, 0xce, 0xe7, 0xe8, 0xca, 0xab, 0xc2, 0x6d, 0x69, 0x92,
	    0x8c, 0xd5, 0x8d, 0x9b, 0x85, 0x44, 0x07, 0xdb, 0x98, 0x19, 0x5b,
	    0xdb, 0xd1, 0x4b, 0x12, 0x3d, 0x9a, 0x56, 0x61, 0x24, 0x4a, 0x29,
	    0x72, 0xe6, 0x1c, 0x39, 0x66, 0xfe, 0x41, 0x62, 0x51, 0x6d, 0x51,
	    0xad, 0x9b, 0x9f, 0xad, 0x30, 0x74, 0x76, 0x0b, 0x4f, 0xa6, 0xbb,
	    0x4d, 0xe6, 0xd5, 0x2b, 0x12, 0x2b, 0x04, 0xd9, 0xce, 0x04, 0x00,
	    0xa0, 0xae, 0x3f, 0x33, 0xdc, 0x46, 0xe9, 0x40, 0xbc, 0x84, 0x9a,
	    0x30, 0xca, 0xfa, 0x61, 0xa5, 0x7c, 0x22, 0xc9, 0xa7, 0x6b, 0x52,
	    0xe4, 0xd5, 0x8a, 0x16, 0x62, 0x58, 0xe5, 0x63, 0xad, 0x9c, 0x7b,
	    0x99, 0xf1, 0x51, 0xb6, 0x7f, 0x1d, 0x33, 0xc3, 0xab, 0xde, 0xe4,
	    0xb2, 0x2e, 0xe5, 0x50, 0xba, 0xf8, 0x03, 0xb2, 0xb1, 0xea, 0xf4,
	    0xaa, 0x8f, 0x4d, 0xe7, 0xe5, 0x98, 0x78, 0x45, 0x63, 0x36, 0xfe,
	    0xf8, 0x6a, 0xf3
	};

	static unsigned char dh2048_g[] = { 0x02 };

	dh = DH_new();
	if (dh == NULL) {
		libarms_log(ARMS_LOG_ESSL, "DH_new() failed");
		return -1;
	}

	dh->p = BN_bin2bn(dh2048_p, sizeof(dh2048_p), NULL);
	dh->g = BN_bin2bn(dh2048_g, sizeof(dh2048_g), NULL);

	if (dh->p == NULL || dh->g == NULL) {
		libarms_log(ARMS_LOG_ESSL, "BN_bin2bn() failed");
		DH_free(dh);
		return -1;
	}

#ifdef ARMS_DEBUG
{
	int ok, codes;
	ok = DH_check(dh, &codes);
	if (!ok)
		ERR_print_errors(bio_err);

	if (codes == 0)
		printf("DH parameters appear to be ok.\n");
	if (codes & DH_CHECK_P_NOT_PRIME)
		printf("p value is not prime\n");
	if (codes & DH_CHECK_P_NOT_SAFE_PRIME)
		printf("p value is not a safe prime\n");
	if (codes & DH_UNABLE_TO_CHECK_GENERATOR)
		printf("unable to check the generator value\n");
	if (codes & DH_NOT_SUITABLE_GENERATOR)
		printf("the g value is not a generator\n");
}
#endif

	SSL_CTX_set_options(ctx, SSL_OP_SINGLE_DH_USE);

	SSL_CTX_set_tmp_dh(ctx, dh);

	DH_free(dh);

	return 0;
}

#ifdef USE_ECDH
static int
arms_ssl_ecdh_curve(SSL_CTX *ctx, const char *name)
{
	int nid;
	EC_KEY *ecdh;

	nid = OBJ_sn2nid(name);
	if (nid == 0) {
		libarms_log(ARMS_LOG_ESSL, "unknown curve name \"%s\"", name);
		return -1;
	}

	ecdh = EC_KEY_new_by_curve_name(nid);
	if (ecdh == NULL) {
		libarms_log(ARMS_LOG_ESSL, "failed to create curve");
		return -1;
	}

	SSL_CTX_set_options(ctx, SSL_OP_SINGLE_ECDH_USE);

	SSL_CTX_set_tmp_ecdh(ctx, ecdh);

	EC_KEY_free(ecdh);

	return 0;
}
#endif

/*
 * SSL_CTX_new wrapper
 */
SSL_CTX *
arms_ssl_ctx_new(int type)
{
#if OPENSSL_VERSION_NUMBER >= 0x00909000L
	const SSL_METHOD *method;
#else
	SSL_METHOD *method;
#endif
	SSL_CTX *ctx;
	long mode;

	switch (type) {
	case ARMS_SSL_SERVER_METHOD:
		method = SSLv23_server_method();
		break;
	case ARMS_SSL_CLIENT_METHOD:
		method = SSLv23_client_method();
		break;
	default:
		/* illegal type */
		return NULL;
		
	}
	ctx = SSL_CTX_new(method);
	if (ctx == NULL)
		return NULL;

	/* disable SSLv2,3*/
	SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
	SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv3);

	if (type == ARMS_SSL_SERVER_METHOD) {
		if (arms_ssl_dhparam(ctx) != 0)
			libarms_log(ARMS_LOG_DEBUG,
			    "can't use DH cipher suites in Push Wait");
#ifdef USE_ECDH
		if (arms_ssl_ecdh_curve(ctx, USE_ECDH) != 0)
			libarms_log(ARMS_LOG_DEBUG,
			    "can't use ECDH cipher suites in Push Wait");
#endif
	}

	/* workaround for SSL_R_BAD_WRITE_RETRY by SSL_write. */
	mode = SSL_CTX_get_mode(ctx);
	/*mode |= SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER;*/
	mode |= SSL_MODE_AUTO_RETRY;
	SSL_CTX_set_mode(ctx, mode);

	return ctx;
}

/*
 * SSL_new wrapper
 */
SSL *
arms_ssl_new(SSL_CTX *ctx)
{
	return SSL_new(ctx);
}

/*
 * SSL_set_fd wrapper
 */
int
arms_ssl_set_fd(SSL *ssl, int fd)
{
	return SSL_set_fd(ssl, fd);
}

/*
 * SSL_CTX_get_cert_store wrapper
 */
X509_STORE *
arms_ssl_ctx_get_cert_store(SSL_CTX *ctx)
{
	return SSL_CTX_get_cert_store(ctx);
}

/*
 * X509_STORE_add_cert wrapper
 */
int
arms_x509_store_add_cert(X509_STORE *store, X509 *cert)
{
	return X509_STORE_add_cert(store, cert);
}

/*
 * SSL_CTX_set_verify_depth wrapper
 */
void
arms_ssl_ctx_set_verify_depth(SSL_CTX *ctx, int depth)
{
	SSL_CTX_set_verify_depth(ctx, depth);
}

/*
 * SSL_use_certificate wrapper
 */
int
arms_ssl_use_certificate(SSL *ssl, X509 *cert)
{
	return SSL_use_certificate(ssl, cert);
}

/*
 * SSL_use_PrivateKey wrapper
 */
int
arms_ssl_use_privatekey(SSL *ssl, EVP_PKEY *key)
{
	return SSL_use_PrivateKey(ssl, key);
}

/*
 * SSL_check_private_key wrapper
 */
int
arms_ssl_check_private_key(SSL *ssl)
{
	return SSL_check_private_key(ssl);
}

/*
 * SSL_set_ex_data wrapper
 */
int
arms_ssl_set_ex_data(SSL *ssl, int idx, void *data)
{
	return SSL_set_ex_data(ssl, idx, data);
}

/*
 * SSL_set_verify wrapper
 */
void
arms_ssl_set_verify(SSL *ssl, int mode,
    int (*callback)(int, X509_STORE_CTX *))
{
	SSL_set_verify(ssl, mode, callback);
}

/*
 * SSL_connect wrapper
 */
int
arms_ssl_connect(SSL *ssl)
{
	int rv;

	rv = SSL_connect(ssl);
	if (rv <= 0) {
		switch(SSL_get_error(ssl, rv)) {
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_NONE:
			return 0;
		case SSL_ERROR_SYSCALL:
			arms_log_ssl_error();
			libarms_log(ARMS_LOG_ESSL,
				    "SSL_connect: syscall errno %d", errno);
			return -1;
		case SSL_ERROR_ZERO_RETURN:
		default:
			arms_log_ssl_error();
			if (rv == 0) {
				rv = -1;
			}
			break;
		}
	}
	return rv;
}

/*
 * SSL_accept wrapper
 */
int
arms_ssl_accept(SSL *ssl)
{
	return SSL_accept(ssl);
}

/*
 * SSL_pending wrapper
 */
int
arms_ssl_pending(SSL *ssl)
{
	return SSL_pending(ssl);
}

/*
 * SSL_read wrapper
 */
int
arms_ssl_read(SSL *ssl, char *buf, int len)
{
	int rv, err;

#ifdef ARMS_DEBUG_SSL_IO
	memset(buf, 0, len);
#endif
	rv = SSL_read(ssl, buf, len);
	if (rv < 0) {
		switch((err = SSL_get_error(ssl, rv))) {
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_NONE:
			/*
			 * in non-blocking mode, negotiation is incomplete.
			 * call select(2) and re-calling it.
			 */
			return 0;
		case SSL_ERROR_SYSCALL:
			arms_log_ssl_error();
			libarms_log(ARMS_LOG_ESSL,
				    "SSL_read: syscall errno %d\n", errno);
			break;
		case SSL_ERROR_ZERO_RETURN:
		default:
			arms_log_ssl_error();
			libarms_log(ARMS_LOG_ESSL,
				    "SSL_read: OpenSSL Connection reset by peer (%d)", err);
			break;
		}
	} else if (rv == 0) {
		/* received shutdown. */
		rv = -1;
	}
#ifdef ARMS_DEBUG_SSL_IO
	else printf("%s: (((%s)))\n", __func__, buf);
#endif
	return rv;
}

/*
 * SSL_write wrapper
 */
int
arms_ssl_write(SSL *ssl, const char *buf, int len)
{
	int rv, err;

	if (len == 0) {
		  libarms_log(ARMS_LOG_DEBUG,
			      "try to write zero bytes. nothing to do.\n");
		  return 0;
	}
	rv = SSL_write(ssl, buf, len);
	if (rv < 0) {
		switch((err = SSL_get_error(ssl, rv))) {
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_NONE:
			/*
			 * in non-blocking mode, negotiation is incomplete.
			 * call select(2) and re-calling it.
			 */
			return 0;
		case SSL_ERROR_SYSCALL:
			arms_log_ssl_error();
			libarms_log(ARMS_LOG_ESSL,
				    "SSL_write: syscall errno %d\n", errno);
			break;
		case SSL_ERROR_ZERO_RETURN:
		default:
			arms_log_ssl_error();
			libarms_log(ARMS_LOG_ESSL,
				    "SSL_write: OpenSSL Connection reset by peer (%d)", err);
			break;
		}
	} else if (rv == 0) {
		/* received shutdown. */
		rv = -1;
	}
#ifdef ARMS_DEBUG_SSL_IO
	else libarms_log(ARMS_LOG_DEBUG, "%s: (((%s)))\n", __func__, buf);
#endif
	return rv;
}

/*
 * SSL_get_error wrapper
 */
int
arms_ssl_get_error(SSL *ssl, int rv)
{
	return SSL_get_error(ssl, rv);
}

/*
 * return current protocol name such as "TLSv1.2"
 */
const char *
arms_ssl_get_protocol_name(SSL *ssl)
{
	const SSL_SESSION *session = SSL_get_session(ssl);
	const char *protocol;

	if (session == NULL)
		return "(error)";

	switch (session->ssl_version) {
#ifdef TLS1_2_VERSION
	case TLS1_2_VERSION:
		protocol = "TLSv1.2";
		break;
#endif
#ifdef TLS1_1_VERSION
	case TLS1_1_VERSION:
		protocol = "TLSv1.1";
		break;
#endif
	case TLS1_VERSION:
		protocol = "TLSv1.0";
		break;
	case SSL3_VERSION:
		protocol = "SSLv3";
		break;
	default:
		protocol = "(unknown)";
		break;
	}

	return protocol;
}

/*
 * SSL_CIPHER_get_name wrapper
 */
const char *
arms_ssl_cipher_get_name(SSL *ssl)
{
	const SSL_SESSION *session = SSL_get_session(ssl);

	if (session == NULL)
		return "(error)";

	return SSL_CIPHER_get_name(session->cipher);
}

/*
 * SSL_shutdown wrapper
 */
void
arms_ssl_shutdown(SSL *ssl)
{
	int i, fd, on;

	/*
	 * switch to non-blocking mode
	 */
	fd = SSL_get_fd(ssl);
	on = 1;
	ioctl(fd, FIONBIO, &on);
	/*
	 * 0... incomplete shutdown.  recall it.
	 * 1... shutdown complete.
	 * -1.. already shutdown by peer.
	 */
	for (i = 0; i < 4; i++) {
		if (SSL_shutdown(ssl) != 0)
			break;
	}
}

/*
 * SSL_free wrapper
 */
void
arms_ssl_free(SSL *ssl)
{
	SSL_free(ssl);
}

/*
 * SSL_CTX_free wrapper
 */
void
arms_ssl_ctx_free(SSL_CTX *ctx)
{
	SSL_CTX_free(ctx);
}

void
arms_ssl_cleanup(void)
{
	CRYPTO_cleanup_all_ex_data();
	ERR_free_strings();
	ERR_remove_state(0);
	EVP_cleanup();
}
